import Image, os, subprocess, operator, logging, argparse, sys, shutil
from splitterbase import SplitterBase

class LowSpectroSplitter(SplitterBase):
    """Wraps all the methods required to split an audio recaptcha into the words using the density at each point on the spectrogram"""

    join_length = 15
    def __init__(self, debug = logging.CRITICAL, threshold = 253, debug_images = False, filename_prefix = "temp/", filename_suffix = ".mp3", include_sinc = False,
            number_of_words = None, noisered_file = True):
        self.logger = logging.getLogger('LowSpectroSplitter')
        logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.logger.setLevel(debug)
        self.debug_images = debug_images
        self.filename_prefix = filename_prefix
        self.filename_suffix = filename_suffix
        self.threshold = threshold
        self.include_sinc = include_sinc
        self.number_of_words = number_of_words
        self.noisered_file = noisered_file

    def split_captcha(self, filename, split_components = [ "audio", "image" ]):
        """Splits a audio captcha and its spectrogram up into the indvidual components for each word.  If only an audio file or only a spectrogram
            image file is needed for each word, split_components can be used to only select the component you need."""
        self.logger.info("Spliting %s" % filename)

        origfile = self.filename_prefix + filename + self.filename_suffix
        fullfile = origfile 
        if self.noisered_file != "":
            fullfile = self.filename_prefix + filename + "_clean" + self.filename_suffix
            shutil.copy(origfile, fullfile)
            self.remove_noise(fullfile, self.noisered_file, extension = self.filename_suffix)

        spectrogram = fullfile[:-len(self.filename_suffix)] + self.filename_suffix 
        options = [ 'sox', fullfile, '-n' ]
        if self.include_sinc:
            options.extend([ 'sinc', '80-160', '-a', '180', '-p', '100' ])
        options.extend([ 'spectrogram', '-o', spectrogram, '-y', '64', '-x', '1000' ])
        p = subprocess.Popen(options, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        p.communicate()
        
        image = Image.open(spectrogram).convert("RGB")
        image,box = self.remove_white_border(image)
        self.image_index = {} #Reset the file naming convention

        if self.debug_images:
            image.save(self.filename_prefix + filename + "_noborder.png", "PNG")

        threshold = self.threshold
        join = self.join_length
        words = self.split_into_words(image, threshold, join)
        if self.number_of_words != None:
            while len(words) < self.number_of_words and join > 5:
                join -= 5
                self.logger.debug("Couldn't find %d words, lowering threshold to %d", self.number_of_words, threshold)
                words = self.split_into_words(image, threshold, join)
            if len(words) < self.number_of_words:
                words = self.split_into_words(image, self.threshold, self.join_length)

        sizex, sizey = image.size
        if "image" in split_components:
            self.get_spectrogram(filename)
            color_image = Image.open(spectrogram).convert("RGB")
            color_image,box = self.remove_white_border(color_image)
            color_sizex,color_sizey = color_image.size
            conversion_rate = float(color_sizex)/float(sizex)
            for i in range(0,len(words)):
                cropped_image = color_image.crop((int(words[i][0] * conversion_rate), 0, int(words[i][1] * conversion_rate), color_sizey))
                cropped_image.save(self.filename_prefix + filename + "_" + str(i) + ".png", "PNG")

        if "audio" in split_components:
            self.split_wav_file(words, sizex, filename)

        return len(words)

    def split_into_words(self, image, threshold, join, minimum_syllable = 3, minimum_word = 4):
        sizex, sizey = image.size
        targety = sizey-2
        pixels = image.load()
        words = []
        inword = False
        start = -1
        for x in range(0,sizex):
            amount = pixels[x,targety][0]
            if not inword and amount > threshold:
                inword = True
                start = x
            elif inword and amount < threshold:
                inword = False
                if x - start >= minimum_syllable:
                    words.append([start,x])
        if inword:
            words.append([start, sizex-1])
        self.logger.info("Found %d words: %s", len(words), words)

        if len(words) == 0:
            self.logger.debug("Splitting failed")
            return []

        new_words = [words[0]]
        for i in range(1, len(words)):
                if words[i][0] - words[i-1][1] < join:
                    new_words[-1][1] = words[i][1]
                else:
                    new_words.append(words[i])
        self.logger.debug("After joining, %d words: %s", len(new_words), new_words)
        if self.number_of_words != None:
            words = self.get_longest(new_words, self.number_of_words) 
        else:
            words = new_words

        self.logger.debug("Best %d words: %s", len(words), words)
        return words

        for word in words:
            x = word[0]
            while x > 0 and pixels[x,targety][0] > 50:
                x -= 1
            word[0] = x
            x = word[1]
            while x < sizex and pixels[x,targety][0] > 50:
                x += 1
            word[1] = x
        self.logger.debug("After expanding %d words: %s", len(words), words)
               
        new_words = [words[0]]
        for i in range(1, len(words)):
                if words[i][0] - words[i-1][1] < join:
                    new_words[-1][1] = words[i][1]
                else:
                    new_words.append(words[i])
        if self.number_of_words != None:
            words = self.get_longest(new_words, self.number_of_words) 
        else:
            words = new_words
        self.logger.debug("After joining, %d words: %s", len(new_words), new_words)

        new_words = []
        for word in words: 
            length = word[1] - word[0]
            if length > minimum_word:
                if word[1] - word[0] < 10: #Stupid six/two fix (There's probably a better way to do this)
                    word[0] -= 3
                    word[1] += 3
                new_words.append(word)
        words = new_words
    

        self.logger.debug("Best %d words: %s", len(words), words)
        return words
    
    def get_longest(self, words, amount):
        """Given a list of lists containing 2 x coordinates, it will return the $amount largest items 
                (largest meaning the most distance between words[i][0] and words[i][1])"""
        lengths = {}
        for i in range(0, len(words)):
            lengths[i] = words[i][1] - words[i][0]
        lengths = sorted(lengths.iteritems(), key=operator.itemgetter(1))
        lengths.reverse()

        indexes = []
        for i in range(0, min(len(lengths), amount)):
            indexes.append(lengths[i][0])
        indexes.sort()

        new_words = []
        for i in indexes:
            new_words.append([words[i][0], words[i][1]])
        return new_words

if __name__ == "__main__":    
    parser = argparse.ArgumentParser(description="Split an audio captcha into the individual words.  For the most of these options, you can mostly leave the defaults unless you feel like experimenting")
    parser.add_argument('filename', help='The spectrogram (from sox) of the audio file to split into individual words')
    parser.add_argument('-d', required=False, action='store_true', help='Debug mode.  Prints out an extra spectrogram with the axis removed and each word with a white border')
    parser.add_argument('-nr', required=False, type=str, default="", help='Use this file to reduce noise before splitting with the lowspect splitter')
    parser.add_argument('-p', required=False, action='store_true', help='Output png files in addition to the wav files (useful for debugging)')
    parser.add_argument('-s', required=False, action='store_true', help='Run the sinc filter before splitting')
    parser.add_argument('-v', required=False, action='store_true', help='Verbose mode')
    parser.add_argument('-vv', required=False, action='store_true', help='Very Verbose mode')

    args = parser.parse_args()
    output_level = logging.CRITICAL
    debug_images = False
    if args.vv:
        output_level = logging.DEBUG
        debug_images = True
    elif args.v:
        output_level = logging.INFO

    dirname, filename = os.path.split(args.filename)
    if dirname != "":
        dirname += "/"
    extension = filename.split(".")[-1]
    filename = filename[:-1-len(extension)]

    captcha_splitter = LowSpectroSplitter(debug = output_level, debug_images = debug_images, filename_prefix = dirname, filename_suffix = "." + extension, 
                                            include_sinc = args.s, noisered_file = args.nr)
    split_components = [ "audio" ]
    if args.p:
        split_components.append("image")
    num_words = captcha_splitter.split_captcha(filename, split_components)
    print num_words



