import Image, os, subprocess, operator, logging, argparse, sys
from splitterbase import SplitterBase

class NucaptchaSplitter(SplitterBase):
    """Wraps all the methods required to split an audio recaptcha into the words using the density at each point on the spectrogram"""

    sample_rate = 11025.0
    start_captcha = 1.25 #The number of seconds in the captcha to start looking for words
    join_length = 20
    def __init__(self, debug = logging.CRITICAL, threshold = 251, debug_images = False, filename_prefix = "temp/", filename_suffix = ".mp3", include_sinc = True,
                    give_both = False):
        self.logger = logging.getLogger('NucaptchaSplitter')
        logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.logger.setLevel(debug)
        self.debug_images = debug_images
        self.filename_prefix = filename_prefix
        self.filename_suffix = filename_suffix
        self.threshold = threshold
        self.include_sinc = include_sinc
        self.give_both = give_both

    def split_captcha(self, filename, split_components = [ "audio", "image" ]):
        """Splits a audio captcha and its spectrogram up into the indvidual components for each word.  If only an audio file or only a spectrogram
            image file is needed for each word, split_components can be used to only select the component you need."""
        self.logger.info("Spliting %s" % filename)

        spectrogram = self.filename_prefix + filename + ".png"
        options = [ 'sox', self.filename_prefix + filename + self.filename_suffix, '-n', 'sinc', '80-160', '-a',
                        '180', '-p', '100', 'spectrogram', '-o', spectrogram, '-y', '64', '-x', '1000' ]
        options = [ 'sox', self.filename_prefix + filename + self.filename_suffix, '-n' ]
        if self.include_sinc:
            options.extend([ 'sinc', '80-160', '-a', '180', '-p', '100' ])
        options.extend([ 'spectrogram', '-o', spectrogram, '-y', '64', '-x', '1000' ])

        p = subprocess.Popen(options, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        p.communicate()
        
        image = Image.open(spectrogram).convert("RGB")
        image,box = self.remove_white_border(image)
        self.image_index = {} #Reset the file naming convention

        if self.debug_images:
            image.save(self.filename_prefix + filename + "_noborder.png", "PNG")

        duration = self.get_duration(self.filename_prefix + filename + self.filename_suffix, True)
        sizex, sizey = image.size
        minx = int(self.start_captcha/duration * sizex)
        words = self.split_into_words(image, self.threshold, minx, self.join_length)

        if "image" in split_components:
            self.get_spectrogram(filename)
            color_image = Image.open(spectrogram).convert("RGB")
            color_image,box = self.remove_white_border(color_image)
            color_sizex,color_sizey = color_image.size
            conversion_rate = float(color_sizex)/float(sizex)
            for i in range(0,len(words)):
                cropped_image = color_image.crop((int(words[i][0] * conversion_rate), 0, int(words[i][1] * conversion_rate), color_sizey))
                cropped_image.save(self.filename_prefix + filename + "_" + str(i) + ".png", "PNG")

        if "audio" in split_components:
            self.split_wav_file(words, sizex, filename, soundsizex = duration * self.sample_rate) #Save some time by passing it to the function since we got it beforehand

        return len(words)

    def split_into_words(self, image, threshold = 253, minx = 0, join = 20):
        sizex, sizey = image.size
        targety = sizey-2
        pixels = image.load()
        words = []
        inword = False
        start = -1
        for x in range(minx,sizex):
            amount = pixels[x,targety][0]
            if not inword and amount > threshold:
                inword = True
                start = x
            elif inword and amount < threshold:
                inword = False
                if x - start >= 5:
                    words.append([start,x])
        if inword:
            words.append([start, sizex-1])
        self.logger.info("Found %d words: %s", len(words), words)

        if len(words) == 0:
            self.logger.debug("Splitting failed")
            return []
            
        new_words = [words[0]]
        for i in range(1, len(words)):
                if words[i][0] - words[i-1][1] < join:
                    new_words[-1][1] = words[i][1]
                else:
                    new_words.append(words[i])
        words = new_words
        self.logger.debug("After joining, %d words: %s", len(words), str(words))

        first_round = []
        instructions_word = -1
        for i in range(0, len(words)):
            if words[i][1] - words[i][0] > 58:
                instructions_word = i
                break
            first_round.append(words[i])
    
        if instructions_word == -1:
            self.logger.debug("Couldn't find instruction word by length, assuming middle word") 
            if len(words) % 2 == 1: #Odd number of words
                instructions_word = (len(words)-1)/2
            else: 
                instructions_word = len(words)/2
            first_round = words[0:instructions_word-1]
        if not self.give_both: #If they only want one set of words, return just the first set
            return first_round

        second_round = words[instructions_word+1:]
        if len(second_round) + 1 == len(first_round): #The first word of the second round was too close to the instructions
            self.logger.debug("First word of second round too close to instructions, looking for it")
            begin_word = -1
            for x in range(words[instructions_word][1] - 10, words[instructions_word][0], -1):
                if pixels[x,targety][0] < threshold:
                    begin_word = x
                    break
            if begin_word != -1:
                second_round.insert(0,[begin_word,words[instructions_word][1]])
                self.logger.debug("Found it at %s", str(second_round[0]))
            else:
                second_round = None
                self.logger.debug("Couldn't find it")

        self.logger.debug("First round %d words: %s", len(first_round), str(first_round))
        self.logger.debug("Second round %d words: %s", len(second_round), str(second_round))

        if second_round != None:
            first_round.extend(second_round)
        return first_round

if __name__ == "__main__":    
    parser = argparse.ArgumentParser(description="Split an audio captcha into the individual words.  For the most of these options, you can mostly leave the defaults unless you feel like experimenting")
    parser.add_argument('filename', help='The spectrogram (from sox) of the audio file to split into individual words')
    parser.add_argument('-d', required=False, action='store_true', help='Debug mode.  Prints out an extra spectrogram with the axis removed and each word with a white border')
    parser.add_argument('-p', required=False, action='store_true', help='Output png files in addition to the wav files (for phash)')
    parser.add_argument('-v', required=False, action='store_true', help='Verbose mode')
    parser.add_argument('-vv', required=False, action='store_true', help='Very Verbose mode')

    args = parser.parse_args()
    output_level = logging.CRITICAL
    debug_images = False
    if args.vv:
        output_level = logging.DEBUG
        debug_images = True
    elif args.v:
        output_level = logging.INFO

    dirname, filename = os.path.split(args.filename)
    if dirname != "":
        dirname += "/"
    extension = filename.split(".")[-1]
    filename = filename[:-1-len(extension)]

    captcha_splitter = NucaptchaSplitter(debug = output_level, debug_images = debug_images, filename_prefix = dirname, filename_suffix = "." + extension, give_both = True)
    split_components = [ "audio" ]
    if args.p:
        split_components.append("image")
    num_words = captcha_splitter.split_captcha(filename, split_components)

