import subprocess, os, shutil, operator, logging, hashlib, random

class CaptchaSolver:
    """A main helper class that wraps all the methods required to defeat recaptcha"""

    tempdir = "temp/"
    number_of_words = 10 #Let it start off with 10, but that's overrided with how many the splitter found
    def __init__(self, solvers, splitter, downloaders, solver_thresholds = [0], masher = None, debug = logging.CRITICAL):
        """Declares a new CaptchaSolver class that will try to solve each word with the given solvers.  If a solver reaches the corresponding value
           for certainness in solver_thresholds, that solver's answer will be selected.  Otherwise it will continue to try other solvers.  The 
           captcha will be split using the given splitter, and all answers will be run through the given word masher if one was given.  The parameter
           debug determines how much information is printed."""
        self.logger = logging.getLogger('CaptchaSolver')
        logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.logger.setLevel(debug)
        
        self.downloaders = downloaders
        self.solver_thresholds = solver_thresholds
        self.solvers = solvers
        self.split_components = [] #Holds a list of all the components needed from splitting
        for solver in self.solvers:
            for split_component in solver.split_components:
                if split_component not in self.split_components:
                    self.split_components.append(split_component)

        self.splitter = splitter
        self.masher = masher
        if not os.path.exists(self.tempdir): #Create the temp folder
            os.mkdir(self.tempdir)


    def make_folders(self, root= "automated/"):
        """Make the necessary folders to save the temp files to (using save_files)"""
        if root[-1] != "/":
            root += "/"
        subfolders = [ "samples", "spectrograms", "solved_clusters", "incorrect/",
                        "incorrect/samples", "incorrect/spectrograms", "incorrect/solved_clusters" ]
        if not os.path.exists(root):
            os.mkdir(root)
        for folder in subfolders:
            if not os.path.exists(root + folder):
                os.mkdir(root + folder)

    def get_next_captcha(self, filename):
        """Sets up the next captcha to be split and solved, using one of the random downloaders.
            Either returns the challenge of the current captcha, or an empty string on an error (most likely tempban)"""
        self.current_downloader = self.downloaders[random.randint(0,len(self.downloaders)-1)]
        challenge, url = self.current_downloader.get_audio_url_and_challenge()
        if challenge != "" and self.current_downloader.get_audio(url, filename):
            return challenge
        else:
            self.current_downloader.fix_error()
            return ""

    def answer_captcha(self, challenge, answer):
        """Submit the captcha to determine if we got it right"""
        success = self.current_downloader.submit_captcha(challenge, answer)
        if success == None:
            self.current_downloader.fix_error()
        return success

    def solve(self, filename):
        """Solves an entire captcha"""
        answer = ""
        split = False
        start_solver = 0
        number_of_words = self.number_of_words
        predMatrices = {}

        self.solvers[0].reset_solver()
        if len(self.solvers[0].split_components) != 0:
            number_of_words = self.splitter.split_captcha(filename, self.split_components)
            split = True
        self.solvers[0].start_solving(filename, number_of_words)
        predMatrices[0] = self.solvers[0].finish_solving(filename)

        #If md5s couldn't find it, just move on to the next solver now.  Rather than wait and move to the next one word at a time (saves overall time)
        if self.solvers[0].classname == "Md5Solver" and predMatrices[0] == self.solvers[0].incorrect_matrix and len(self.solvers) > 1: 
            start_solver += 1
            self.solvers[1].reset_solver()
            if len(self.solvers[1].split_components) != 0:
                number_of_words = self.splitter.split_captcha(filename, self.split_components)
                split = True
            self.solvers[1].start_solving(filename, number_of_words)
            predMatrices[1] = self.solvers[1].finish_solving(filename)

        for i in range(0, number_of_words):
            word = ""
            pwords=[]
            foundThreshold = False
            skipWord = False

            for j in range(start_solver, len(self.solvers)):
                if not j in predMatrices or i not in predMatrices[j]:
                    skipWord = True
                    break
                word = max(predMatrices[j][i].iteritems(), key=operator.itemgetter(1))[0]

                if word != "":
                    self.logger.info("%s (solver %d) choosing %s for word %d with %.4f", self.solvers[j].classname, j, word, i, predMatrices[j][i][word])
                else: #The Md5Solver can't always choose a word
                    self.logger.info("%s (solver %d) couldn't choose an answer for word %d", self.solvers[j].classname, j, i)
                  
                pwords.append([predMatrices[j][i][word] * self.solvers[j].accuracy_multiplier,word])

                if self.solver_thresholds[j] < predMatrices[j][i][word]: 
                    foundThreshold = True
                    break

                if j + 1 < len(self.solvers):
                    solver = self.solvers[j+1]
                    if len(solver.split_components) != 0 and not split:
                        number_of_words = self.splitter.split_captcha(filename, self.split_components)
                        split = True

                    solver.reset_solver()
                    solver.start_solving_word(solver.filename_prefix + filename + "_" + str(i) + solver.filename_suffix)
                    if j+1 not in predMatrices:
                        predMatrices[j+1] = {}
                    predMatrices[j+1][i] = solver.finish_solving_word(0)

            if not foundThreshold and not skipWord:
                certainty, word = max(pwords)
                
                best_solver = -1
                for solver_index in range(0, len(pwords)):
                    if certainty == pwords[solver_index][0] and word == pwords[solver_index][1]:
                        best_solver = solver_index
                        break
                if len(pwords) > 1:
                    self.logger.info("%s %d solved with the best match of %s with %f certainty", self.solvers[best_solver].classname,
                                    best_solver, word, certainty)
                best_pred_matrix = predMatrices[best_solver][i]
            elif not skipWord:
                self.logger.info("%s %d decided %s with certainity of %f (which is better than threshold %f)", self.solvers[j].classname, 
                                    j, word, predMatrices[j][i][word], self.solver_thresholds[j])
                best_pred_matrix = predMatrices[j][i]
            
            if answer != "":
                answer += " "

            if self.masher != None: 
                answer += self.masher.mash_word(word, best_pred_matrix)
            else:
                answer += word

        return answer

    def save_files(self, success, answer, filename, path = "automated/"):
        """Moves all the files from the captcha solving process to the correct directory"""
        if path[-1] != '/':
            path += "/"
        if answer.find(" ") != -1:
            answer_parts = answer.split(" ")
        else:
            answer_parts = answer

        if not success:
           path += "incorrect/"

        #Save all the files accordingly
        if os.path.exists(self.tempdir + filename + ".mp3"):
            shutil.copy(self.tempdir + filename + ".mp3", path + "samples/" + filename + ".mp3-" + answer.replace(" ", "_") + ".mp3")
        if os.path.exists(self.tempdir + filename + ".wav"):
            shutil.copy(self.tempdir + filename + ".wav", path + "samples/" + filename + ".wav-" + answer.replace(" ", "_") + ".wav")
        
        if os.path.exists(self.tempdir + filename + ".png"):
            shutil.move(self.tempdir + filename + ".png", path + "spectrograms/" + filename + ".png")
        if os.path.exists(self.tempdir + filename + ".flac"):
            shutil.move(self.tempdir + filename + ".flac", path + "samples/" + filename + ".flac")

        if os.path.exists(self.tempdir + filename + "_noborder.png"):
            shutil.move(self.tempdir + filename + "_noborder.png", path + "spectrograms/" + filename + "_noborder.png")
        if os.path.exists(self.tempdir + filename + "_debug.png"):
            shutil.move(self.tempdir + filename + "_debug.png", path + "spectrograms/" + filename + "_debug.png")
        for j in range(0,len(answer_parts)):
            splitfile = filename + "_" + str(j) + ".png"
            debugsplitfile = filename + "_" + str(j) + "before.png"
            wavsplitfile = filename + "_" + str(j) + ".wav"
            debugwavsplitfile = filename + "_" + str(j) + "before.wav"
            destdir = path + "solved_clusters/" + answer_parts[j]
            if not os.path.exists(destdir):
                os.mkdir(destdir)
            if os.path.exists(self.tempdir + splitfile):
                shutil.move(self.tempdir + splitfile, destdir + "/" + splitfile)
            if os.path.exists(self.tempdir + debugsplitfile):
                shutil.move(self.tempdir + debugsplitfile, destdir + "/" + debugsplitfile)
            if os.path.exists(self.tempdir + wavsplitfile):
                shutil.move(self.tempdir + wavsplitfile, destdir + "/" + wavsplitfile)
            if os.path.exists(self.tempdir + debugwavsplitfile):
                shutil.move(self.tempdir + debugwavsplitfile, destdir + "/" + debugwavsplitfile)
        #Delete any files we didn't want to keep
        self.delete_files(filename)

    def delete_files(self, filename):
        """Deletes all files downloaded / generated while solving the captcha"""
        files_to_delete = [self.tempdir + filename + ".mp3", self.tempdir + filename + ".png", 
                           self.tempdir + filename + "_noborder.png", self.tempdir + filename + "_debug.png",
                           self.tempdir + filename + ".flac", self.tempdir + filename + ".wav",
                           self.tempdir + "cookies" + str(os.getpid()) ]
        for i in range(0,10):
            files_to_delete.append(self.tempdir + filename + "_" + str(i) + ".png")
            files_to_delete.append(self.tempdir + filename + "_" + str(i) + "before.png")
            files_to_delete.append(self.tempdir + filename + "_" + str(i) + ".wav")
            files_to_delete.append(self.tempdir + filename + "_" + str(i) + "before.wav")

        for delete_file in files_to_delete:
            if os.path.exists(delete_file):
                os.remove(delete_file)

    def feedback(self, success, mashed_answer, filename):
        """Records the MD5 sum if correct and lets the solvers know whether they got it right 
            so they can track internal accuracies"""
        if success:
            if os.path.exists(self.tempdir + filename + ".mp3"):
                self.record_md5sum(self.tempdir + filename + ".mp3", mashed_answer)
            elif os.path.exists(self.tempdir + filename + ".wav"):
                self.record_md5sum(self.tempdir + filename + ".wav", mashed_answer)
        if len(self.solvers) == 1: #We really only care about tracking the solvers internal accuracy when using a single solver
            self.solvers[0].feedback(success)

    def record_md5sum(self, filename, mashed_answer):
        """Adds the md5 of the filename and answer to the new_md5s file"""
        #Remove comment tag (you can't add randomness to a part we can strip)
        p = subprocess.Popen(["id3tool", "--set-note=", filename], stdin = subprocess.PIPE, stdout = subprocess.PIPE)
        p.communicate()

        m = hashlib.md5()
        captchaFile = open(filename, "rb")
        m.update(captchaFile.read())
        captchaFile.close()
        
        captchaFile = open("new_md5s", "a")
        captchaFile.write(m.hexdigest() + " " + mashed_answer.replace(" ", "_") + "\n")
        captchaFile.close()

